/**
 * Column constructor
 *
 * `sqlName` is the name of the column in the database table, e.g. `user_id`.
 * `validator` is a function that checks the columns value before inserts and
 *             updates and also serves to determine the columns type.
 *             When the value doesn't match the expected type or shape, any
 *             error that function raises will be catched and wrapped in a
 *             QueryBuilderValidationError that contains additional context
 *             information such as table and column name where the error
 *             occured.
 * `fromJson` is a function that is called to turn a projected json value into
 *            the columns type, e.g. turn an iso-date-string into a Javascript
 *            Date. We need this when using `selectAsJson` or
 *            `selectAsJsonAgg` and the column is of type Date.
 */
export declare function column(sqlName: string): Column<unknown>;
export declare function column<T>(sqlName: string, validator: (value: unknown) => T, // checks/converts the datatype when inserting a column
fromJson?: (value: unknown) => T): Column<T>;
declare type EnumObject = {
    [key: string]: string | number;
};
/**
 * A column of a table
 */
export declare class Column<T> {
    private columnValue;
    private name;
    private isNullable?;
    private fromJson?;
    private isPrimaryKey?;
    constructor(params: {
        name: string;
        columnValue: (value: unknown) => T;
        fromJson?: (value: unknown) => T;
        isPrimaryKey?: true;
        isNullable?: true;
    });
    private copy;
    /**
     * Mark this column as being the sole or part of the tables primary key.
     *
     * Knowing that a column is a primary key is required to generate correct
     * group-by clauses for json_agg (selectAsJsonAgg) projections.
     */
    primary(): Column<T>;
    /**
     * Mark this column has having a default value.
     *
     * Columns with defaults can be ommitted in insert queries.
     * `nullable` values always have null as the default.
     */
    default(): Column<T & {
        __typesafeQueryBuilderHasDefault?: true;
    }>;
    /**
     * Make this column nullable.
     *
     * That means it can hold `null` and also uses null as its default.
     */
    null(): Column<T | null>;
    private checkThatColumnValueIsIdentity;
    /**
     * Map this column to a typescript number that must be an integer.
     *
     * Column values being integers is only checked when inserting or updating
     * data.
     */
    integer(): Column<number>;
    /**
     * Map this column to a string.
     *
     * postgres types: TEXT, VARCHAR
     */
    string(): Column<string>;
    /**
     * Map this column to a boolean.
     *
     * postgres type: BOOLEAN
     */
    boolean(): Column<boolean>;
    /**
     * Map this column to a date.
     *
     * postgres types: TIMESTAMPTZ
     */
    date(): Column<Date>;
    /**
     * Map this column to an json object.
     *
     * Validator should be function that validates the type of the incoming
     * data. Validator is called before inserting or updating and stringifies
     * any data before passing it to postgres.
     *
     * postgres types: JSON, JSONB
     */
    json<J>(validator: (data: unknown) => J): Column<J>;
    /**
     * Map a string union to a
     *
     * postgres types: TEXT, VARCHAR
     */
    stringUnion<A extends string>(a: A): Column<A>;
    stringUnion<A extends string, B extends string>(a: A, b: B): Column<A | B>;
    stringUnion<A extends string, B extends string, C extends string>(a: A, b: B, c: C): Column<A | B | C>;
    stringUnion<A extends string, B extends string, C extends string, D extends string>(a: A, b: B, c: C, d: D): Column<A | B | C | D>;
    stringUnion<A extends string, B extends string, C extends string, D extends string, E extends string>(a: A, b: B, c: C, d: D, e: E): Column<A | B | C | D | E>;
    stringUnion<A extends string, B extends string, C extends string, D extends string, E extends string, F extends string>(a: A, b: B, c: C, d: D, e: E, f: F): Column<A | B | C | D | E | F>;
    stringUnion<A extends string, B extends string, C extends string, D extends string, E extends string, F extends string, G extends string>(a: A, b: B, c: C, d: D, e: E, f: F, g: G): Column<A | B | C | D | E | F | G>;
    enum<T extends EnumObject, S extends keyof T>(enumObject: T): Column<T[S]>;
}
/**
 * Internal interface that maps exactly on a Column
 *
 * Basically just turns the private fields onto public fields and throws away
 * the generic type information. When generating sql queries, we don't need
 * that.
 */
export interface ColumnImplementation {
    columnValue: (value: unknown) => any;
    fromJson?: (value: unknown) => any;
    isNullable?: true;
    isPrimaryKey?: true;
    name: string;
    copy: (params: {
        name: string;
    }) => ColumnImplementation;
}
/**
 * Return the implementation of a column.
 *
 * The implementation is used to create queries at runtime.
 * The column type is used provide type checking at compile type.
 */
export declare function getColumnImplementation(column: Column<any>): ColumnImplementation;
export {};
